(* environment.ml *)

open Ast;;

module VariableMap = Map.Make(String);;
module FunctionMap = Map.Make(String);;

exception EmptyEnvironmentError;;
exception NameAlreadyBoundError of string;;
exception VariableNotFound of string;;
exception VariableAlreadyDeclared;; 
exception AlreadyDeclaredError;; 
exception FunctionNotDefinedError;;
exception ReservedWordError of string;;
  
type func_info  = {
	id : string; 
        gpu : bool;
	return : datatype; 
	args : datatype list;
	arg_names: string list;
	_blocksize : int;
}

(* Record indicating what the current environment keeps track of *)
type env = {
    var_stack: datatype VariableMap.t list; 
    func_return_type_map: func_info FunctionMap.t; 
    current_function: string; 
    on_gpu: bool;
    gfunc_list: fdecl list;	(* we need to save the body of the
                                   gfunc so we can do kernel string
                                   generation *)
    num_array_returns : int;
    var_array_sizes : int VariableMap.t; 
  } 

(* Types that can be returned by the generator as it modifies either
   the text of the generated code changes the environment as it is
   parsing the file or passes fuunctions to edit both code and
   environment either in the existing scope or in a new scope *)
type source = 
  | Text of string
  | Env of (env -> env) 
  | Generator of (env -> (string * env))
  | NewScope of (env -> (string * env))

(* Create initializes an empty record for environment *)
let add_default_func f_id fmap = 
  let f_info = 
    { 
      id = f_id; 
      gpu = false;
      return = Int; 
      args = [];
      arg_names = [];
      _blocksize = -1; 
    }
  in 
  (FunctionMap.add f_id f_info fmap)

let init_func_map = 
  let time_start_fmap = 
    let printf_fmap = 
      add_default_func "printf" FunctionMap.empty 
    in 
    add_default_func "time_start" printf_fmap 
  in 
  add_default_func "time_end" time_start_fmap 
  
let create = 
  {
    var_stack = VariableMap.empty::[];
    func_return_type_map = init_func_map; 
    current_function = ""; (* TODO maybe this needs a better convention *)
    on_gpu = false;
    gfunc_list = [];
    num_array_returns = 0;
    var_array_sizes = VariableMap.empty;   
  }

(* Update gives a new env record with updated values of the record *)
let update v_stack f_map curr_f gpu g_list num_arr_ret  v_a_s= 
  { 
    var_stack = v_stack; 
    func_return_type_map = f_map;
    current_function = curr_f; 
    on_gpu = gpu; 
    gfunc_list = g_list;
    num_array_returns = num_arr_ret;
    var_array_sizes = v_a_s; 
  }

(* Functions that let us modify only one variable in environment at a
   time *)
let update_only_scope new_scope env = 
  update new_scope env.func_return_type_map env.current_function env.on_gpu env.gfunc_list env.num_array_returns env.var_array_sizes

let update_only_func new_func env = 
  update env.var_stack new_func env.current_function env.on_gpu env.gfunc_list env.num_array_returns env.var_array_sizes

let update_curr_func new_curr_func env = 
  update env.var_stack env.func_return_type_map new_curr_func env.on_gpu env.gfunc_list env.num_array_returns env.var_array_sizes

let update_on_gpu gpu env = 
  update env.var_stack env.func_return_type_map env.current_function gpu env.gfunc_list env.num_array_returns env.var_array_sizes

let update_gfunc_list g_list env = 
  update env.var_stack env.func_return_type_map env.current_function env.on_gpu g_list env.num_array_returns env.var_array_sizes

let increment_num_array_rets env =
  update env.var_stack env.func_return_type_map env.current_function env.on_gpu env.gfunc_list (env.num_array_returns + 1) env.var_array_sizes

let add_array_size id size env = 
  update env.var_stack env.func_return_type_map env.current_function env.on_gpu env.gfunc_list env.num_array_returns  (VariableMap.add id size env.var_array_sizes)

(* Checks all scopes to see if variable has been declared *)
let is_var_in_scope id env = 
  let rec check_scopes scope_stack =
    let check_level scope_level =
      VariableMap.mem id scope_level
    in
    match scope_stack with 
    | [] -> false
    | [scope_level] ->
       if env.on_gpu then false
       else check_level scope_level
    | scope_level :: other_scope_levels ->
       if check_level scope_level then true
       else check_scopes other_scope_levels 
  in check_scopes env.var_stack 

(* Checks all scopes to find variable and returns type if found
 * Raises exception if not found *)
let typeof id env = 
  let rec check_scopes scope_stack = 
    match scope_stack with 
    | [] -> raise (VariableNotFound id) 
    | scope_level :: other_scope_levels ->
       if VariableMap.mem id scope_level then 
         VariableMap.find id scope_level 
       else 
         check_scopes other_scope_levels 
  in check_scopes env.var_stack

(* Adds a variable to the topmost level of the variable stack. Does
   not check if variable is already in stack do elseware raises error
   if stack is an empty list returns an updated environment with an
   updated scope stack *)
let update_scope_add_var id datatype env = 
  let old_scope, scope_tail = 
    ( match env.var_stack with 
      | old_scope :: scope_tail -> old_scope, scope_tail
      | [] -> raise EmptyEnvironmentError ) in 
  let new_scope = VariableMap.add id datatype old_scope in
  update_only_scope (new_scope::scope_tail) env    
		    
(* Handles adding a variable to the current environment. 

   Takes a tuple of
   - id : string 
   - datatype 
   - (text: string, env : env)
  
   and either raises 
   - EmptyError
   - AlreadyDeclaredError

   or it adds the variable to the current top scope with
   update_scope_add_var and returns a updated env *)
let add_var id datatype env = 
    match env.var_stack with 
      | [] -> raise EmptyEnvironmentError 
      | scope_level :: scope_tail -> 
		if VariableMap.mem id scope_level then 
			raise VariableAlreadyDeclared
		else 
		update_scope_add_var id datatype env

let get_arr_size id env = 
    if VariableMap.mem id env.var_array_sizes then 
      (VariableMap.find id env.var_array_sizes)
    else 
     raise VariableAlreadyDeclared 

(* adds a new empty variableMap to the top of var stack used to enter
   a subscope takes an env, returns an updated env *)
let push_scope env =  
    update_only_scope (VariableMap.empty::env.var_stack) env

(* removes a variableMap from the top of var stack used to enter a
   subscope takes an env, returns an updated env *)
let pop_scope env = 
   match env.var_stack with 
    | popped_scope::other_scopes -> 
        update_only_scope other_scopes env
    | [] -> raise EmptyEnvironmentError 

(* --------------------------------------------------------- *)
(* The following methods deal with handling the function map *)
(* --------------------------------------------------------- *)

(* Takes a function id and checks if the function is defined
   returns bool *) 
let is_func_declared id env = 
   FunctionMap.mem id env.func_return_type_map 

(* Takes a function id and either returns the mapped function info 
   or raises an undefined function error *)
let get_func_info id env = 
  if is_func_declared id env then 
    FunctionMap.find id env.func_return_type_map
  else 
    raise FunctionNotDefinedError

let get_func_args id env =
    (get_func_info id env).args

(* Inserts a new function to the function map and updates environment
   or raises a AlreadyDeclaredError *)
let add_func id finfo env = 
  if is_func_declared id env then 
    raise VariableAlreadyDeclared
  else
    update_only_func (FunctionMap.add id finfo env.func_return_type_map) env	   
       
(* Returns the datatype of a function or 
   raises a undefined error if the function is not defined *) 
let return_typeof_func id env = 
    let f_info = 
	get_func_info id env in 
    f_info.return     

let id_is_gfunc id env =
  let rec name_in_gfunc_list gfunc_list =
    match gfunc_list with
      [] -> false
    | gfunc :: other_gfuncs ->
       if id = gfunc.fname then true
       else name_in_gfunc_list other_gfuncs
  in
  name_in_gfunc_list env.gfunc_list

let rec check_gfunc_name_in_list glist gfunc_fdecl = 
  match glist with 
    [] -> false 
  | gfunc_fdecl :: rest_of_gfuncs -> if gfunc_fdecl.fname = gfunc_fdecl.fname then true
                                     else (check_gfunc_name_in_list rest_of_gfuncs gfunc_fdecl) 

let is_gfunc_declared gfunc_fdecl env = 
  if check_gfunc_name_in_list env.gfunc_list gfunc_fdecl then 
    raise (AlreadyDeclaredError) 
  else if is_func_declared gfunc_fdecl.fname env then 
    raise (AlreadyDeclaredError) 
  else false   

let add_gfunc gfunc_fdecl env = 
  if (is_gfunc_declared gfunc_fdecl env) then 
    raise (AlreadyDeclaredError)
  else if (gfunc_fdecl.fname = "main") then
    raise (ReservedWordError("a gfunc cannot be the main method"))
  else
    update_gfunc_list (gfunc_fdecl :: env.gfunc_list) env

let quote_and_strip_newline str =
  let len = String.length str in
  let last_char s = s.[(len - 1)] in
  if str = "" then ""
  else
    match last_char str with
      '\n' -> (String.sub str 0 (len - 2))
    | _    -> str

(* This is the running loop of the codegen step. First, f is a
   function that takes a (text, environment) tuple and matches it with
   a component which is either 

   * some new text - gets appended to existing text
   * some modified environment - replaces existing environment
   * some function named gen applied to the same scope

   gen is applied to the existing environment and gen returns a tuple
   of (string, env). The text gets appended to the existing text, and
   the new env replaces the old env.

   Note that gen can be a function also with arguments given to it
   then this function f is applied to component in the list of
   components always returning an updated (text, env) tuple that is
   passed to the next component in the list *)
let append init_env components =
  let f (text, env) component =
    match component with 
    | Text(str) -> if env.on_gpu then
		     text ^ (quote_and_strip_newline str), env
		   else
		     text ^ str, env
    | Env (env_gen) -> let new_env = env_gen env in 
		       text, new_env
    | Generator(gen) -> let new_str, new_env = gen env in 
			text ^ new_str, new_env
    | NewScope(gen) -> let new_str, new_env = gen (push_scope env) in 
		       text ^ new_str, pop_scope new_env in
  List.fold_left f("", init_env) components
;;
